
'''

Solve a Poisson transmission problem 

- \Delta p_top    = f_top      in \Omega_top    := (0,1)x(0.5, 1)
- \Delta p_bottom = f_bottom   in \Omega_bottom := (0,1)x(0, 0.5)

                        p_top - p_bottom = jump1                on interface
         \nabla p_top \cdot n - \nabla p_bottom \cdot n = jump2 on interface 

with a hybrid mixed FEM using the "element_boundary" programming style of NGSolve.

The exact solution is: 

    p_top    = sin(pi*x)*sin(pi*y) 
and 
    p_bottom = 8*x*(1-x)*y*(1-y)

'''

# The "element_boundary" version  gives a wrong solution:
# I guess that the problem is related to the orientation of the normal vector on the interface.
# Trying to solve without static condensation throws the error: "UmfpackInverse: Numeric factorization failed"

from ngsolve import *
from netgen.geom2d import SplineGeometry

from numpy import pi

ngsglobals.msg_level = 1


# Different sets of: Exact solution + source terms 

# Case 1

p_top    = sin(pi*x)*sin(pi*y)
p_bottom = 8*x*(1-x)*y*(1-y)
ftop     = 2*pi**2*p_top
fbottom  = 16*y*(1-y) + 16*x*(1-x)
gradp_top    = CoefficientFunction( (pi*cos(pi*x)*sin(pi*y), pi*sin(pi*x)*cos(pi*y)) )
gradp_bottom = CoefficientFunction( ( 8*(1-2*x)*y*(1-y), 8*x*(1-x)*(1-2*y) ) )

# Case 2

""" p_top    = sin(pi*x)*sin(pi*y)
p_bottom = cos(pi*x)*cos(pi*y)
ftop     = 2*pi**2*p_top
fbottom  = 2*pi**2*p_bottom
gradp_top    = CoefficientFunction( (pi*cos(pi*x)*sin(pi*y), pi*sin(pi*x)*cos(pi*y)) )
gradp_bottom    = -CoefficientFunction( (pi*sin(pi*x)*cos(pi*y), pi*cos(pi*x)*sin(pi*y)) ) """

# Case 3

""" p_top = cos(pi*x)*cos(pi*y)
p_bottom = -p_top
ftop = 2*pi**2*p_top
fbottom = -ftop
gradp_top = -CoefficientFunction( (pi*sin(pi*x)*cos(pi*y), pi*cos(pi*x)*sin(pi*y)) )
gradp_bottom = -gradp_top """

# Case 4

""" p_top = atan((y+0.5)/(x+0.5))
p_bottom = -p_top
ftop = 0
fbottom = -ftop
gradp_top    = CoefficientFunction( (- (y+0.5)/( (x+0.5)**2 + (y+0.5)**2 ) , (x+0.5)/((x+0.5)**2 + (y+0.5)**2)) )
gradp_bottom = -gradp_top """

# jumps on interface

jump1 = p_top - p_bottom
jump2 = gradp_top - gradp_bottom

# Geometry and mesh

def topBottom():
    geometry = SplineGeometry()
    pnts     = [ (0, 0), (1, 0),  (1, 0.5), (1, 1), (0, 1), (0, 0.5)]
    pnums    = [geometry.AppendPoint(*p) for p in pnts]
    # start-point, end-point, boundary-condition, domain on left side, domain on right side:
    lines = [ (pnums[0], pnums[1], "gammaBottom",  2, 0),
              (pnums[1], pnums[2], "gammaBottom",  2, 0),
              (pnums[2], pnums[5], "interface",   2, 1),
              (pnums[5], pnums[0], "gammaBottom",  2, 0),
              #(pnums[5], pnums[2], "interface",   1, 2), 
              (pnums[2], pnums[3], "gammaTop",     1, 0),
              (pnums[3], pnums[4], "gammaTop",     1, 0),
              (pnums[4], pnums[5], "gammaTop",     1, 0)]
    for p1,p2,bc,left,right in lines:
        geometry.Append( ["line", p1, p2], bc=bc, leftdomain=left, rightdomain=right)

    geometry.SetMaterial(1, "top")
    geometry.SetMaterial(2, "bottom")
        
    return geometry


mesh = Mesh(topBottom().GenerateMesh (maxh=0.2))

n = specialcf.normal(mesh.dim)

pExact = CoefficientFunction([p_top if mat == "top" else p_bottom for mat in mesh.GetMaterials()])

# Finite element spaces

order = 4

V1 = HDiv(mesh, order = order,   definedon="top",    discontinuous=True)
M1 = L2(mesh,   order = order-1, definedon="top")
V2 = HDiv(mesh, order = order,   definedon="bottom", discontinuous=True)
M2 = L2(mesh,   order = order-1, definedon="bottom")
L  = FacetFESpace(mesh, order=order, dirichlet="gammaTop|gammaBottom")

M = L2(mesh,  order = order-1) # Used to project exact solution

X = FESpace([V1, V2, M1, M2, L])
u1, u2, p1, p2, l  = X.TrialFunction()
v1, v2, q1, q2, m  = X.TestFunction()


# The bilinear  form

a  = BilinearForm(X, symmetric = True, condense= True)
a += u1*v1*dx('top') + u2*v2*dx('bottom')
a += div(u1)*q1*dx('top')    + div(v1)*p1*dx('top')
a += div(u2)*q2*dx('bottom') + div(v2)*p2*dx('bottom')

a +=  -l*(v1*n)*dx('top',    element_boundary=True)  - m*(u1*n)*dx('top',    element_boundary=True)
a +=  -l*(v2*n)*dx('bottom', element_boundary=True)  - m*(u2*n)*dx('bottom', element_boundary=True)

# The right-hand side

f  = LinearForm(X)
f += -ftop * q1 * dx('top') - fbottom * q2 * dx('bottom') #Source term
f += p_top*(v1.Trace()*n)*ds(definedon=mesh.Boundaries("gammaTop")) #Dririchlet BC (natural)
f += p_bottom*(v2.Trace()*n)*ds(definedon=mesh.Boundaries("gammaBottom")) #Dirichlet BC (natural)

f += 0.5*jump1 * (v1.Trace()*n + v2.Trace()*n) * ds(definedon=mesh.Boundaries("interface")) # First transmission condition
f += -(jump2*n) *m.Trace()*ds(definedon=mesh.Boundaries("interface")) # Second transmission condition


# Solving the problem (with static condensation)

c = Preconditioner(a, type = "direct")
a.Assemble()
f.Assemble()

gf = GridFunction(X)
gf.components[4].Set(p_top, BND, definedon=mesh.Boundaries("gammaTop"))
gf.components[4].Set(p_bottom, BND, definedon=mesh.Boundaries("gammaBottom"))

f.vec.data += a.harmonic_extension_trans * f.vec
solvers.CG(mat=a.mat, pre=c.mat, rhs=f.vec, sol=gf.vec, initialize=False)
#gf.vec.data += a.mat.Inverse(freedofs=X.FreeDofs(), inverse="umfpack")*f
gf.vec.data += a.harmonic_extension * gf.vec
gf.vec.data += a.inner_solve * f.vec

# Drawing the solution

gfp1 = gf.components[2]
gfp2 = gf.components[3]



glob = GridFunction(M, name="aprox")
glob.Set(gfp1+gfp2) # The global solution

# L2-error

upe = GridFunction(M, name="exact")
upe.Set(pExact)


err = sqrt(Integrate((upe-glob)*(upe-glob),mesh))
print(err)

Draw(upe)
Draw(glob)